/***************************************************************************\
  "table.hpp" - Hash Table (Phase C)
  Revision 13; 10 Nov 2002 - 15 March 2004
  Copyright (C) 2002-2004  Kevin P. Albrecht

  References:
    Dragon pp433-437
    MCD 2.1.11; 2.1.11.3
\***************************************************************************/

#ifndef TABLE_HPP
#define TABLE_HPP

/////////////////////////
// DEFINES //////////////
/////////////////////////

// External Defines //
// GLOBAL_VALUE_STRING_SIZE

// This value (known as "m") should be a prime number so that
//   if "n" is the average number of symbols in the symbol table,
//   "n / m" is approximately <= 2.  See (Aho pp433-437).
// According to Grune, a good "m" is 1008 (MCD 2.1.11.3)
#define TABLE_HASH_ARRAY_SIZE 17

// This value (known as "alpha") is multiplied by each character
//   in the symbol's key before they are added together.
//   According to (Aho p437), alpha=1 is the least efficient
//   and alpha=65599 is best for 32-bit integers.
// According to Grune, a good "alpha" is 613 (MCD 2.1.11.3)
#define TABLE_HASH_ALPHA 65599

/////////////////////////
// INCLUDES /////////////
/////////////////////////

// Standard //

// Local //
#include "symbol.hpp"

/////////////////////////
// DECLARATIONS /////////
/////////////////////////

// Bucket //
class Bucket
{
  public:
    Bucket ();

    Symbol *search (char *aKey);
    Symbol *enter (char *aKey);

    Symbol *head;
    Symbol *tail;
};

// Table //
class Table
{
  public:
    Table ();
    ~Table ();

    Symbol *search (char *aKey);
    Symbol *enter (char *aKey);

    static int hash (char *as);

    Bucket *array[TABLE_HASH_ARRAY_SIZE]; // hash array
};

/////////////////////////
// DEFINITIONS //////////
/////////////////////////

Bucket::Bucket ()
{
  head = tail = NULL;
}

//// Returns pointer to the Symbol with the matching
//// symbol key, or NULL if not found
Symbol *Bucket::search (char *aKey)
{
  Symbol *temp = head;

  // Loop to find the matching symbol key //
  while (temp)
  {
    if (k_same (temp->key, aKey))
    {
      return temp;
    }
    else
    {
      temp = temp->next;
    }
  }

  // Not found; return NULL //
  return NULL;
}

//// Returns pointer to the Symbol; If not found, it will be created
Symbol *Bucket::enter (char *aKey)
{
  Symbol *temp = search (aKey);

  // Not found; create and enter it //
  if (temp == NULL)
  {
    temp = new Symbol (aKey);

    // First element //
    if (head == NULL)
    {
      head = tail = temp;
    }
    // Other elements //
    else
    {
      tail->next = temp;
      tail = temp;
    }
  }

  // Found //
  else
  {
  }

  // Return //
  return temp;
}

Table::Table ()
{
  // Initialize the hash array to NULL //
  for (int i = 0; i < TABLE_HASH_ARRAY_SIZE; i++)
  {
    array[i] = NULL;
  }
}

Table::~Table ()
{
  // Delete each element in array //
  for (int i = 0; i < TABLE_HASH_ARRAY_SIZE; i++)
  {
    if (array[i]) delete array[i];
  }
}

//// Returns pointer to the node, or NULL if not found
Symbol *Table::search (char *aKey)
{
  Bucket *pBucket = array[hash (aKey)];;

  // There is no Bucket at that hash; Not found //
  if (pBucket == NULL)
  {
    return NULL;
  }

  // Bucket found at that hash //
  else
  {
    return pBucket->search (aKey);
  }
}

//// Returns pointer to the node; If not found, it will be created
Symbol *Table::enter (char *aKey)
{
  int nHash = hash (aKey);

  // There is no Bucket at that hash; Create Bucket //
  if (array[nHash] == NULL)
  {
    array[nHash] =  new Bucket ();
  }

  return array[nHash]->enter (aKey);
}

//// hash function for the "Alpha" algorithm (Aho p435)
int Table::hash (char *as)
{
  int i;
  int len;
  int total;

  len = strlen (as);
  total = 0;

  for (i = 0; i < len; i++)
  {
    total += as[i] * TABLE_HASH_ALPHA;
  }

  return (total % TABLE_HASH_ARRAY_SIZE);
}

#endif
